use axum::{
    extract::Json,
    response,
    routing::{get, post},
    Router,
};
use battlesnake::types::{
    simulation::Board,
    wire::{Request, Response},
    Direction,
};
use log::{debug, error, info, warn};
use rand::prelude::*;
use serde::Serialize;
use tokio::{
    net::TcpListener,
    time::{Duration, Instant},
};

#[tokio::main]
async fn main() {
    env_logger::init();

    debug!("Creating routes");
    let app = Router::new()
        .route("/", get(info))
        .route("/start", post(start))
        .route("/move", post(get_move))
        .route("/end", post(end));

    debug!("Creating listener");
    let listener = TcpListener::bind("0.0.0.0:8000").await.unwrap();
    debug!("Starting server");
    axum::serve(listener, app).await.unwrap();
}

async fn info() -> response::Json<Info> {
    info!("got info request");
    response::Json(Info {
        apiversion: "1",
        author: "der-informatiker",
        color: "#00FFEE",
        head: "smart-caterpillar",
        tail: "mouse",
        version: env!("CARGO_PKG_VERSION"),
    })
}

#[derive(Debug, Clone, Serialize)]
struct Info {
    apiversion: &'static str,
    author: &'static str,
    color: &'static str,
    head: &'static str,
    tail: &'static str,
    version: &'static str,
}

async fn start(request: Json<Request>) {
    let board = Board::from(&*request);
    info!("got start request: {board}");
}

async fn get_move(request: Json<Request>) -> response::Json<Response> {
    let start = Instant::now();
    let board = Board::from(&*request);
    let id = board.get_id(&request.you.id).unwrap_or_else(|| {
        error!("My id is not in the simulation board");
        0
    });
    info!("got move request: {board}");
    let actions = board.valid_actions(id).collect::<Vec<_>>();
    if actions.len() <= 1 {
        info!(
            "only one possible action. Fast forwarding {:?}",
            actions.first()
        );
        return response::Json(Response {
            direction: actions.first().copied().unwrap_or(Direction::Up),
            shout: None,
        });
    }
    info!("valid actions: {:?}", actions);

    let mut action_data = [(0, 0); 4];
    let mut total_simulations = 0;
    let base_turns = board.turn();
    while start.elapsed() < Duration::from_millis(250) {
        let mut board = board.clone();
        let action = *actions.choose(&mut thread_rng()).unwrap_or(&Direction::Up);
        board.next_turn(&[(id, action)]);
        let turns = board.simulate_random(|board| {
            if board.num_snakes() == 0
                || board.turn() > base_turns + u32::from(request.you.length) * 3
            {
                Some(board.turn())
            } else {
                None
            }
        });
        let action_data = &mut action_data[usize::from(action)];
        action_data.0 += turns - base_turns;
        action_data.1 += 1;
        total_simulations += 1;
    }
    debug!("action data: {action_data:?}");

    let action = actions.into_iter().max_by_key(|action| {
        let action_data = action_data[usize::from(*action)];
        (action_data.0 / action_data.1, action_data.0 % action_data.1)
    });

    if let Some(action) = action {
        let action_data = action_data[usize::from(action)];
        let avg_turns = action_data.0 / action_data.1;
        info!("found action {action:?} after {total_simulations} simulations with an average of {avg_turns} turns.");
    } else {
        warn!("unable to find a valid action");
    }
    info!("chose {action:?}");
    response::Json(Response {
        direction: action.unwrap_or(Direction::Up),
        shout: None,
    })
}

async fn end(request: Json<Request>) {
    let board = Board::from(&*request);
    info!("got end request: {board}");
}
